# to test this:
# pytest -s -m thread_safe_attributes
import re
import inspect
from threading import RLock
from collections import namedtuple

FrameData = namedtuple('FrameData', [
  'filename', 
  'line_number',
  'function_name',
  'lines',
  'index'])

class ThreadSafeAttribute:

  def __init__(self, initial_value=None):
    self._initial_value = initial_value
    self._is_atomic = True
    self._lock = RLock()
    self._value = initial_value

  def is_not_atomic(self, previous_line):
    is_not_atomic = True
    # search for '+=', '-=' ... '<<=', '**='
    is_not_atomic &= re.search(r'([+-/*@^&|<>%]=)|([/<>*]{2}=)', previous_line) is not None
    return is_not_atomic

  def request_for_lock(self, previous_line):
    request_for_lock = False
    # search for _, _lock = ...
    if '_lock' in previous_line:
      request_for_lock |= re.search(r'_, _lock[ ]+=', previous_line) is not None
    return request_for_lock

  def __get__(self, instance, owner):
    #print("get acquiring lock")
    self._lock.acquire(blocking=True)
    self._is_atomic = True
    previous_frame = inspect.currentframe().f_back
    fdata = FrameData(*inspect.getframeinfo(previous_frame))
    previous_line = fdata.lines[0]
    if self.is_not_atomic(previous_line):
      #print('{} not atomic'.format(fdata.lines[0]))
      self._is_atomic = False
    else:
      #print('{} is atomic'.format(fdata.lines[0]))
      #print("get releasing lock")
      self._lock.release()
    if self.request_for_lock(previous_line):
      #print("providing lock")
      return self._value, self._lock
    else:
      return self._value

  def __set__(self, instance, value):
    if self._is_atomic:
      #print("set aquiring lock")
      self._lock.acquire(blocking=True)
    else:
      #print("set continuing non atomic operation")
      pass
    self._value = value
    self._is_atomic = True
    #print("set releasing lock")
    self._lock.release()

class MetaThreadSafeAttributes(type):

  def __init__(cls, *args, **kwargs):
    '''Build thread safe attributes'''
    if hasattr(cls, '_attributes'):
      for name in list(set(cls._attributes)):
        setattr(cls, name, ThreadSafeAttribute(initial_value=0))
    super().__init__(*args, **kwargs)



from miros.event import Event
from miros.event import signals
from miros.event import return_status

from miros.hsm import pp
from miros.hsm import spy_on
from miros.hsm import HsmWithQueues

from miros.activeobject import ActiveObject
from miros.activeobject import Factory
from miros.hsm import stripped
from miros.hsm import InstrumentedHsmEventProcessor
from miros.thread_safe_attributes import MetaThreadSafeAttributes
from miros.activeobject import FactoryWithAttributes
from miros.activeobject import ActiveObjectWithAttributes
from miros.activeobject import ThreadSafeAttributes



from collections  import deque
from miros.hsm import HsmEventProcessor, Attribute


class ForeignHsm():
  '''
  Provides an object that can be filled with the trace/spy information coming
  from another statechart on a different host.
  '''
  def __init__(self):
    self.foreign = Attribute()
    self.foreign.spy = deque(maxlen=HsmEventProcessor.SPY_RING_BUFFER_SIZE)
    self.foreign.trace = deque(maxlen=HsmEventProcessor.TRC_RING_BUFFER_SIZE)

  def clear_spy(self):
    self.foreign.spy.clear()

  def clear_trace(self):
    self.foreign.trace.clear()

  def trace(self):
    strace = ""
    for tr in self.foreign.trace:
      strace += tr
      strace += "\n"
    return strace

  def spy(self):
    return list(self.foreign.spy)

  def append_to_spy(self, item):
    self.foreign.spy.append(item)

  def append_to_trace(self, item):
    self.foreign.trace.append(item)




"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],

    name='miros',
    py_modules=['miros'],

    # https://packaging.python.org/en/latest/single_source_version.html
    version='4.1.4',

    description='A statechart library for Python',
    long_description=long_description,

    # The project's main homepage.
    url='https://github.com/aleph2c/miros',

    # Author details
    author='Scott Volk',
    author_email='scottvolk@gmail.com',

    # Choose your license
    license='Python 3.5 license',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: Python Software Foundation License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    # What does your project relate to?
    keywords='hsm HSM statechart hierarchical state machine statemachine miros',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['doc', 'docs', 'test', 'experiment', 'examples', 'venv', 'plan']),

)
